@@ -1,14 +1,12 @@
 #!/usr/bin/env python2
 # vim:fileencoding=UTF-8:ts=4:sw=4:sta:et:sts=4:fdm=marker:ai
-from __future__ import (unicode_literals, division, absolute_import,
-                        print_function)
+from __future__ import absolute_import, division, print_function, unicode_literals
 
 __license__   = 'GPL v3'
 __copyright__ = '2012, Kovid Goyal <kovid at kovidgoyal.net>'
 __docformat__ = 'restructuredtext en'
 
-import json, os
-from future_builtins import map
+import json, os, numbers
 from math import floor
 from collections import defaultdict
 
@@ -24,6 +22,9 @@ from calibre.ebooks.pdf.render.common import (inch, cm, mm, pica, cicero,
                                               didot, PAPER_SIZES, current_log)
 from calibre.ebooks.pdf.render.engine import PdfDevice
 from calibre.ptempfile import PersistentTemporaryFile
+from calibre.utils.resources import load_hyphenator_dicts
+from calibre.utils.monotonic import monotonic
+from polyglot.builtins import iteritems, itervalues, map, unicode_type
 
 
 def get_page_size(opts, for_comic=False):  # {{{
@@ -90,10 +91,10 @@ class Page(QWebPage):  # {{{
         self.longjs_counter = 0
 
     def javaScriptConsoleMessage(self, msg, lineno, msgid):
-        self.log.debug(u'JS:', unicode(msg))
+        self.log.debug(u'JS:', unicode_type(msg))
 
     def javaScriptAlert(self, frame, msg):
-        self.log(unicode(msg))
+        self.log(unicode_type(msg))
 
     @pyqtSlot(result=bool)
     def shouldInterruptJavaScript(self):
@@ -127,19 +128,19 @@ def draw_image_page(page_rect, painter, p, preserve_aspect_ratio=True):
 
 class PDFWriter(QObject):
 
-    @pyqtSlot(result=unicode)
+    @pyqtSlot(result=unicode_type)
     def title(self):
         return self.doc_title
 
-    @pyqtSlot(result=unicode)
+    @pyqtSlot(result=unicode_type)
     def author(self):
         return self.doc_author
 
-    @pyqtSlot(result=unicode)
+    @pyqtSlot(result=unicode_type)
     def section(self):
         return self.current_section
 
-    @pyqtSlot(result=unicode)
+    @pyqtSlot(result=unicode_type)
     def tl_section(self):
         return self.current_tl_section
 
@@ -149,6 +150,7 @@ class PDFWriter(QObject):
         QObject.__init__(self)
 
         self.logger = self.log = log
+        self.mathjax_dir = P('mathjax', allow_user_override=False)
         current_log(log)
         self.opts = opts
         self.cover_data = cover_data
@@ -159,10 +161,15 @@ class PDFWriter(QObject):
         self.view = QWebView()
         self.page = Page(opts, self.log)
         self.view.setPage(self.page)
-        self.view.setRenderHints(QPainter.Antialiasing|
-                    QPainter.TextAntialiasing|QPainter.SmoothPixmapTransform)
+        self.view.setRenderHints(QPainter.Antialiasing|QPainter.TextAntialiasing|QPainter.SmoothPixmapTransform)
         self.view.loadFinished.connect(self.render_html,
                 type=Qt.QueuedConnection)
+        self.view.loadProgress.connect(self.load_progress)
+        self.ignore_failure = None
+        self.hang_check_timer = t = QTimer(self)
+        t.timeout.connect(self.hang_check)
+        t.setInterval(1000)
+
         for x in (Qt.Horizontal, Qt.Vertical):
             self.view.page().mainFrame().setScrollBarPolicy(x,
                     Qt.ScrollBarAlwaysOff)
@@ -189,7 +196,7 @@ class PDFWriter(QObject):
                              xdpi=xdpi, ydpi=ydpi, errors=self.log.error,
                              debug=self.log.debug, compress=not
                              opts.uncompressed_pdf, opts=opts,
-                             mark_links=opts.pdf_mark_links)
+                             mark_links=opts.pdf_mark_links, page_margins=(ml, mr, mt, mb))
         self.footer = opts.pdf_footer_template
         if self.footer:
             self.footer = self.footer.strip()
@@ -214,6 +221,10 @@ class PDFWriter(QObject):
         self.margin_top, self.margin_bottom = map(lambda x:int(floor(x)), (mt, mb))
 
         self.painter = QPainter(self.doc)
+        try:
+            self.book_language = pdf_metadata.mi.languages[0]
+        except Exception:
+            self.book_language = 'eng'
         self.doc.set_metadata(title=pdf_metadata.title,
                               author=pdf_metadata.author,
                               tags=pdf_metadata.tags, mi=pdf_metadata.mi)
@@ -250,9 +261,10 @@ class PDFWriter(QObject):
             raise Exception('PDF Output failed, see log for details')
 
     def render_inline_toc(self):
+        evaljs = self.view.page().mainFrame().evaluateJavaScript
         self.rendered_inline_toc = True
         from calibre.ebooks.pdf.render.toc import toc_as_html
-        raw = toc_as_html(self.toc, self.doc, self.opts)
+        raw = toc_as_html(self.toc, self.doc, self.opts, evaljs)
         pt = PersistentTemporaryFile('_pdf_itoc.htm')
         pt.write(raw)
         pt.close()
@@ -274,13 +286,28 @@ class PDFWriter(QObject):
             self.loop.exit(1)
 
     def render_next(self):
-        item = unicode(self.render_queue.pop(0))
+        item = unicode_type(self.render_queue.pop(0))
 
         self.logger.debug('Processing %s...' % item)
         self.current_item = item
         load_html(item, self.view)
+        self.last_load_progress_at = monotonic()
+        self.hang_check_timer.start()
+
+    def load_progress(self, progress):
+        self.last_load_progress_at = monotonic()
+
+    def hang_check(self):
+        if monotonic() - self.last_load_progress_at > 60:
+            self.log.warn('Timed out waiting for %s to render' % self.current_item)
+            self.ignore_failure = self.current_item
+            self.view.stop()
 
     def render_html(self, ok):
+        self.hang_check_timer.stop()
+        if self.ignore_failure == self.current_item:
+            ok = True
+        self.ignore_failure = None
         if ok:
             try:
                 self.do_paged_render()
@@ -290,7 +317,7 @@ class PDFWriter(QObject):
                 return
         else:
             # The document is so corrupt that we can't render the page.
-            self.logger.error('Document cannot be rendered.')
+            self.logger.error('Document %s cannot be rendered.' % self.current_item)
             self.loop.exit(1)
             return
         done = self.total_items - len(self.render_queue)
@@ -304,7 +331,7 @@ class PDFWriter(QObject):
 
     def load_mathjax(self):
         evaljs = self.view.page().mainFrame().evaluateJavaScript
-        mjpath = P(u'viewer/mathjax').replace(os.sep, '/')
+        mjpath = self.mathjax_dir.replace(os.sep, '/')
         if iswindows:
             mjpath = u'/' + mjpath
         if bool(evaljs('''
@@ -314,8 +341,21 @@ class PDFWriter(QObject):
             self.log.debug('Math present, loading MathJax')
             while not bool(evaljs('mathjax.math_loaded')):
                 self.loop.processEvents(self.loop.ExcludeUserInputEvents)
+            # give the MathJax fonts time to load
+            for i in range(5):
+                self.loop.processEvents(self.loop.ExcludeUserInputEvents)
             evaljs('document.getElementById("MathJax_Message").style.display="none";')
 
+    def load_header_footer_images(self):
+        from calibre.utils.monotonic import monotonic
+        evaljs = self.view.page().mainFrame().evaluateJavaScript
+        st = monotonic()
+        while not evaljs('paged_display.header_footer_images_loaded()'):
+            self.loop.processEvents(self.loop.ExcludeUserInputEvents)
+            if monotonic() - st > 5:
+                self.log.warn('Header and footer images have not loaded in 5 seconds, ignoring')
+                break
+
     def get_sections(self, anchor_map, only_top_level=False):
         sections = defaultdict(list)
         ci = os.path.abspath(os.path.normcase(self.current_item))
@@ -335,35 +375,91 @@ class PDFWriter(QObject):
 
         return sections
 
+    def hyphenate(self, evaljs):
+        evaljs(u'''\
+        Hyphenator.config(
+            {
+            'minwordlength'    : 6,
+            // 'hyphenchar'     : '|',
+            'displaytogglebox' : false,
+            'remoteloading'    : false,
+            'doframes'         : true,
+            'defaultlanguage'  : 'en',
+            'storagetype'      : 'session',
+            'onerrorhandler'   : function (e) {
+                                    console.log(e);
+                                }
+            });
+        Hyphenator.hyphenate(document.body, "%s");
+        ''' % self.hyphenate_lang
+        )
+
+    def convert_page_margins(self, doc_margins):
+        ans = [0, 0, 0, 0]
+
+        def convert(name, idx, vertical=True):
+            m = doc_margins.get(name)
+            if m is None:
+                ans[idx] = getattr(self.doc.engine, '{}_margin'.format(name))
+            else:
+                ans[idx] = m
+
+        convert('left', 0, False), convert('top', 1), convert('right', 2, False), convert('bottom', 3)
+        return ans
+
     def do_paged_render(self):
         if self.paged_js is None:
             import uuid
             from calibre.utils.resources import compiled_coffeescript as cc
-            self.paged_js =  cc('ebooks.oeb.display.utils')
-            self.paged_js += cc('ebooks.oeb.display.indexing')
-            self.paged_js += cc('ebooks.oeb.display.paged')
-            self.paged_js += cc('ebooks.oeb.display.mathjax')
-            self.hf_uuid = str(uuid.uuid4()).replace('-', '')
+            self.paged_js =  cc('ebooks.oeb.display.utils').decode('utf-8')
+            self.paged_js += cc('ebooks.oeb.display.indexing').decode('utf-8')
+            self.paged_js += cc('ebooks.oeb.display.paged').decode('utf-8')
+            self.paged_js += cc('ebooks.oeb.display.mathjax').decode('utf-8')
+            if self.opts.pdf_hyphenate:
+                self.paged_js += P('viewer/hyphenate/Hyphenator.js', data=True).decode('utf-8')
+                hjs, self.hyphenate_lang = load_hyphenator_dicts({}, self.book_language)
+                self.paged_js += hjs
+            self.hf_uuid = unicode_type(uuid.uuid4()).replace('-', '')
 
         self.view.page().mainFrame().addToJavaScriptWindowObject("py_bridge", self)
         self.view.page().longjs_counter = 0
         evaljs = self.view.page().mainFrame().evaluateJavaScript
         evaljs(self.paged_js)
         self.load_mathjax()
+        if self.opts.pdf_hyphenate:
+            self.hyphenate(evaljs)
 
-        amap = evaljs('''
+        margin_top, margin_bottom = self.margin_top, self.margin_bottom
+        page_margins = None
+        if self.opts.pdf_use_document_margins:
+            doc_margins = evaljs('document.documentElement.getAttribute("data-calibre-pdf-output-page-margins")')
+            try:
+                doc_margins = json.loads(doc_margins)
+            except Exception:
+                doc_margins = None
+            if doc_margins and isinstance(doc_margins, dict):
+                doc_margins = {k:float(v) for k, v in iteritems(doc_margins) if isinstance(v, numbers.Number) and k in {'right', 'top', 'left', 'bottom'}}
+                if doc_margins:
+                    margin_top = margin_bottom = 0
+                    page_margins = self.convert_page_margins(doc_margins)
+
+        amap = json.loads(evaljs('''
         document.body.style.backgroundColor = "white";
+        // Qt WebKit cannot handle opacity with the Pdf backend
+        s = document.createElement('style');
+        s.textContent = '* {opacity: 1 !important}';
+        document.documentElement.appendChild(s);
         paged_display.set_geometry(1, %d, %d, %d);
         paged_display.layout();
         paged_display.fit_images();
         ret = book_indexing.all_links_and_anchors();
         window.scrollTo(0, 0); // This is needed as getting anchor positions could have caused the viewport to scroll
-        ret;
-        '''%(self.margin_top, 0, self.margin_bottom))
+        JSON.stringify(ret);
+        '''%(margin_top, 0, margin_bottom)))
 
         if not isinstance(amap, dict):
             amap = {'links':[], 'anchors':{}}  # Some javascript error occurred
-        for val in amap['anchors'].itervalues():
+        for val in itervalues(amap['anchors']):
             if isinstance(val, dict) and 'column' in val:
                 val['column'] = int(val['column'])
         for href, val in amap['links']:
@@ -390,14 +486,19 @@ class PDFWriter(QObject):
             if idx is not None:
                 setattr(self, attr, sections[idx][0])
 
+        from calibre.ebooks.pdf.render.toc import calculate_page_number
+
         while True:
             set_section(col, sections, 'current_section')
             set_section(col, tl_sections, 'current_tl_section')
-            self.doc.init_page()
+            self.doc.init_page(page_margins)
+            num = calculate_page_number(self.current_page_num, self.opts.pdf_page_number_map, evaljs)
             if self.header or self.footer:
-                evaljs('paged_display.update_header_footer(%d)'%self.current_page_num)
+                if evaljs('paged_display.update_header_footer(%d)'%num) is True:
+                    self.load_header_footer_images()
+
             self.painter.save()
-            mf.render(self.painter)
+            mf.render(self.painter, mf.ContentsLayer)
             self.painter.restore()
             try:
                 nsl = int(evaljs('paged_display.next_screen_location()'))
@@ -443,7 +544,6 @@ class ImagePDFWriter(object):
                               tags=pdf_metadata.tags, mi=pdf_metadata.mi)
         self.doc_title = pdf_metadata.title
         self.doc_author = pdf_metadata.author
-        page_rect = QRect(*self.doc.full_page_rect)
 
         for imgpath in items:
             self.log.debug('Processing %s...' % imgpath)
@@ -452,7 +552,7 @@ class ImagePDFWriter(object):
             with lopen(imgpath, 'rb') as f:
                 if not p.loadFromData(f.read()):
                     raise ValueError('Could not read image from: {}'.format(imgpath))
-            draw_image_page(page_rect,
+            draw_image_page(QRect(*self.doc.full_page_rect),
                     self.painter, p,
                     preserve_aspect_ratio=True)
             self.doc.end_page()
